#include <fcntl.h>
#include <ndbm.h>
#include "alt_database.h"

#ifndef COPY
#define COPY(src, dest, dtype) *dest = *((dtype *) src);
#endif

#define swapl(longword) (alt_convert ? htonl(longword) : (longword))

static FILE *dir_f, *user_f, *summary_f;
static DBM *room_dbm;
static summary_list *summary_cache;
static int max_messages;
static int alt_convert = FALSE;

bb_rid Alt_last_rid;
bb_gid Alt_last_gid;
bb_uid Alt_last_uid;

#define DATA_ROOM_ID 'i'
#define DATA_UID 'u'
#define DATA_RID 'r'
#define DATA_GID 'g'
#define DATA_GROUP 'G'

typedef struct {char type; bb_rid rid; bb_uid uid;} room_id;
typedef struct {char type; bb_rid rid; long message;} room_message;
typedef struct {char type; bb_gid gid;} key_group;
typedef struct {char type; char id[ID_SIZE];} data_key_id;
typedef struct {char type; char group[GROUP_SIZE];} data_key_gid;
typedef struct {char type; char name[ROOM_SIZE];} data_key_rid;

static char *rid_flag = "< last rid >", *gid_flag = "< last gid >",
  *uid_flag = "< last uid >";

int alt_database_init(int convert)
{
  alt_convert = convert;
  dir_f = fopen(DIRECTORY, "r");
  user_f = fopen(USER, "r");
  summary_f = fopen(SUMMARY, "r");
  room_dbm = dbm_open(ROOM, O_RDONLY, 0666);
  Alt_last_rid = alt_get_rid(rid_flag);
  Alt_last_gid = alt_get_gid(gid_flag);
  Alt_last_uid = alt_get_uid(uid_flag);
  if(Alt_last_rid >= 0) {
    summary_cache = (summary_list *)
      malloc((Alt_last_rid+1) * sizeof(summary_list));
    bzero((char *) summary_cache, (Alt_last_rid+1) * sizeof(summary_list));
  } else {
    summary_cache = NULL;
  }
  max_messages = 2048; /*** this won't last forever ***/
  return (dir_f && user_f && summary_f && room_dbm);
}

void alt_database_shutdown()
{
  fclose(dir_f);
  fclose(user_f);
  fclose(summary_f);
  dbm_close(room_dbm);
  if(summary_cache) {
    free(summary_cache);
    summary_cache = NULL;
  }
}

int alt_get_user_entry(user_entry *user_info, bb_uid uid)
{
  if(uid < 0) return FALSE;
  fseek(user_f, uid * sizeof(user_entry), 0);
  if(fread(user_info, sizeof(user_entry), 1, user_f) != 1) return FALSE;
  if(alt_convert) convert_user_entry(user_info);
  if(user_info->uses == 0) return FALSE;
  return TRUE;
}

int alt_get_directory_entry(directory_entry *directory_info, bb_rid rid)
{
  if(rid < 0) return FALSE;
  fseek(dir_f, rid * sizeof(directory_entry), 0);

  if(fread(directory_info, sizeof(directory_entry), 1, dir_f) != 1) return 0;
  if(alt_convert) convert_directory_entry(directory_info);
  return directory_info->next_message_number > 0;
}

int alt_get_room_user_entry(room_user_entry *room_user_info,
			    bb_rid rid, bb_uid uid)
{
  datum key, content;
  room_id room_id;

  bzero(&room_id, sizeof(room_id));
  room_id.type = DATA_ROOM_ID;
  room_id.rid = swapl(rid);
  room_id.uid = swapl(uid);
  key.dptr = (char *) &room_id;
  key.dsize = sizeof(room_id);
  content = dbm_fetch(room_dbm, key);
  if(content.dptr == NULL) return FALSE;
  COPY(content.dptr, room_user_info, room_user_entry);
  if(alt_convert) convert_room_user_entry(room_user_info);
  return TRUE;
}

bb_uid alt_get_uid(char *id)
{
  datum key, content;
  data_key_id key_id;
  bb_uid uid;

  bzero(&key_id, sizeof(key_id));
  key_id.type = DATA_UID;
  downcase(key_id.id, id);
  key.dptr = (char *) &key_id;
  key.dsize = sizeof(key_id);
  content = dbm_fetch(room_dbm, key);
  if(content.dptr == NULL) return -1;
  COPY(content.dptr, &uid, bb_uid);
  return swapl(uid);
}

bb_rid alt_get_rid(char *name)
{
  datum key, content;
  data_key_rid key_rid;
  bb_rid rid;

  bzero(&key_rid, sizeof(key_rid));
  key_rid.type = DATA_RID;
  downcase(key_rid.name, name);
  key.dptr = (char *) &key_rid;
  key.dsize = sizeof(key_rid);
  content = dbm_fetch(room_dbm, key);
  if(content.dptr == NULL) return -1;
  COPY(content.dptr, &rid, bb_rid);
  return swapl(rid);
}

bb_gid alt_get_gid(char *group)
{
  datum key, content;
  data_key_gid key_gid;
  bb_gid g;

  bzero(&key_gid, sizeof(key_gid));
  key_gid.type = DATA_GID;
  downcase(key_gid.group, group);
  key.dptr = (char *) &key_gid;
  key.dsize = sizeof(key_gid);
  content = dbm_fetch(room_dbm, key);
  if(content.dptr == NULL) return -1;
  COPY(content.dptr, &g, bb_gid);
  return swapl(g);
}

int alt_get_group_entry(group_entry *group_info, bb_gid gid)
{
  datum key, content;
  key_group key_group;

  bzero(&key_group, sizeof(key_group));
  key_group.type = DATA_GROUP;
  key_group.gid = swapl(gid);
  key.dptr = (char *) &key_group;
  key.dsize = sizeof(key_group);
  content = dbm_fetch(room_dbm, key);
  if(content.dptr == NULL) return FALSE;
  COPY(content.dptr, group_info, group_entry);
  if(alt_convert) convert_group_entry(group_info);
  return TRUE;
}

summary_list *alt_get_summary(bb_rid rid)
{
  summary_entry *summary;
  int count;

  if(summary_cache[rid].version) return &summary_cache[rid];

  fseek(summary_f,
	rid * (sizeof(int) + max_messages * sizeof(summary_entry)),
	0);
  if(fread(&count, sizeof(count), 1, summary_f) != 1) count = 0;
  if(count > max_messages) {
    syslog(LOG_ERR, "get_summary(%d): invalid count: %d\n", count);
    return NULL;
  }
  summary_cache[rid].saved_version = 1;
  if(count > 0) {
    int n;

    summary = (summary_entry *) malloc(count * sizeof(summary_entry));
    if(!summary) {
      syslog(LOG_ERR, "get_summary(%d): malloc: %m\n", rid);
      return NULL;
    }
    n = fread(summary, sizeof(summary_entry), count, summary_f);
    if(n != count) {
      syslog(LOG_ERR, "get_summary(%d): fread: %m\n", rid);
      free(summary);
      return NULL;
    }
  } else {
    summary = NULL;
  }
  summary_cache[rid].count = count;
  summary_cache[rid].list = summary;
  summary_cache[rid].version = 1;
  if(alt_convert) convert_summary_list(&summary_cache[rid]);
  return &summary_cache[rid];
}

char *alt_get_id(bb_uid uid)
{
  user_entry user_info;
  
  return alt_get_user_entry(&user_info, uid) ? user_info.id : "******";
}

#define flipl(a) (a = htonl(a))

void convert_user_entry(user_entry *x)
{
  int i;
  flipl(x->last_access_time);
  flipl(x->uses);
  flipl(x->posts);
  flipl(x->connect_time);
  flipl(x->cutoff_priority);
  for(i = 0; i < MAX_GROUPS; i++)
    flipl(x->group_priority[i]);
}

void convert_directory_entry(directory_entry *x)
{
  flipl(x->owner_uid);
  flipl(x->creation_time);
  flipl(x->default_expiration);
  flipl(x->next_message_number);
  flipl(x->first_message_number);
  flipl(x->room_offset);
}

void convert_room_user_entry(room_user_entry *x)
{
  flipl(x->last_message_read);
  flipl(x->user_offset);
}

void convert_group_entry(group_entry *x)
{
  flipl(x->owner);
  flipl(x->group_offset);
  flipl(x->max_expiration);
}

void convert_summary_list(summary_list *summary)
{
  summary_entry *entry;
  message_entry *message_info;

  for(entry = summary->list;
      entry < summary->list + summary->count;
      entry++) {
    flipl(entry->number);
    message_info = &(entry->header);
    flipl(message_info->date);
    flipl(message_info->expiration);
    flipl(message_info->size);
    flipl(message_info->lines);
    flipl(message_info->poster_uid);
  }
}

